import hashlib
from typing import cast, Any, BinaryIO

def feed_file(h: Any, f: BinaryIO) -> None:
    """Feed data read from an open file into the hashlib instance."""

    while True:
        data = f.read(65536)
        if len(data) == 0:
            # end of file
            break
        h.update(data)

def feed_file_path(h: Any, path: str) -> None:
    """Feed data read from a file (to be opened by this function) into the hashlib instance."""

    with open(path, 'rb') as f:
        feed_file(h, f)

def file_digest(algorithm: Any, path: str) -> str:
    """Calculate the digest of a file and return it in hexadecimal notation."""

    h = algorithm()
    feed_file_path(h, path)
    return cast(str, h.hexdigest())

def guess_digest_algorithm(digest: str) -> Any:
    l = len(digest)
    if l == 32:
        return hashlib.md5
    elif l == 40:
        return hashlib.sha1
    elif l == 64:
        return hashlib.sha256
    elif l == 128:
        return hashlib.sha512
    else:
        return None

def verify_file_digest(path: str, expected_digest: str) -> bool:
    """Verify the digest of a file, and return True if the digest matches with the given expected digest."""

    algorithm = guess_digest_algorithm(expected_digest)
    assert(algorithm is not None)
    return file_digest(algorithm, path) == expected_digest



import subprocess
from typing import Union

from .toolchain import AnyToolchain

def run_quilt(toolchain: AnyToolchain, cwd: str, patches_path: str, *args: str) -> None:
    env = dict(toolchain.env)
    env['QUILT_PATCHES'] = patches_path
    subprocess.check_call(['quilt'] + list(args), cwd=cwd, env=env)

def push_all(toolchain: AnyToolchain, src_path: str, patches_path: str) -> None:
    run_quilt(toolchain, src_path, patches_path, 'push', '-a')



import os, shutil
import re
from typing import cast, BinaryIO, Optional, Sequence, Union

from build.download import download_basename, download_and_verify
from build.tar import untar
from build.quilt import push_all
from .toolchain import AnyToolchain

class Project:
    def __init__(self, url: Union[str, Sequence[str]], md5: str, installed: str,
                 name: Optional[str]=None, version: Optional[str]=None,
                 base: Optional[str]=None,
                 patches: Optional[str]=None,
                 edits=None,
                 use_cxx: bool=False):
        if base is None:
            basename = download_basename(url)
            m = re.match(r'^(.+)\.(tar(\.(gz|bz2|xz|lzma))?|zip)$', basename)
            if not m: raise RuntimeError('Could not identify tarball name: ' + basename)
            self.base = m.group(1)
        else:
            self.base = base

        if name is None or version is None:
            m = re.match(r'^([-\w]+)-(\d[\d.]*[a-z]?[\d.]*(?:-(?:alpha|beta)\d+)?)(\+.*)?$', self.base)
            if not m: raise RuntimeError('Could not identify tarball name: ' + self.base)
            if name is None: name = m.group(1)
            if version is None: version = m.group(2)

        self.name = name
        self.version = version

        self.url = url
        self.md5 = md5
        self.installed = installed

        if patches is not None:
            srcdir = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
            patches = os.path.join(srcdir, patches)
        self.patches = patches
        self.edits = edits
        self.use_cxx = use_cxx

    def download(self, toolchain: AnyToolchain) -> str:
        return download_and_verify(self.url, self.md5, toolchain.tarball_path)

    def is_installed(self, toolchain: AnyToolchain) -> bool:
        tarball = self.download(toolchain)
        installed = os.path.join(toolchain.install_prefix, self.installed)
        tarball_mtime = os.path.getmtime(tarball)
        try:
            return os.path.getmtime(installed) >= tarball_mtime
        except FileNotFoundError:
            return False

    def unpack(self, toolchain: AnyToolchain, out_of_tree: bool=True) -> str:
        if out_of_tree:
            parent_path = toolchain.src_path
        else:
            parent_path = toolchain.build_path
        path = untar(self.download(toolchain), parent_path, self.base,
                     lazy=out_of_tree and self.patches is None)
        if self.patches is not None:
            push_all(toolchain, path, self.patches)

        if self.edits is not None:
            for filename, function in self.edits.items():
                with open(os.path.join(path, filename), 'r+t') as f:
                    old_data = f.read()
                    new_data = function(old_data)
                    f.seek(0)
                    f.truncate(0)
                    f.write(new_data)

        return path

    def make_build_path(self, toolchain: AnyToolchain, lazy: bool=False) -> str:
        path = os.path.join(toolchain.build_path, self.base)
        if lazy and os.path.isdir(path):
            return path
        try:
            shutil.rmtree(path)
        except FileNotFoundError:
            pass
        os.makedirs(path, exist_ok=True)
        return path

    def build(self, toolchain: AnyToolchain) -> None:
        self._build(toolchain)



import os.path, subprocess

from build.project import Project

class FfmpegProject(Project):
    def __init__(self, url, md5, installed, configure_args=[],
                 cppflags='',
                 **kwargs):
        Project.__init__(self, url, md5, installed, **kwargs)
        self.configure_args = configure_args
        self.cppflags = cppflags

    def _build(self, toolchain):
        src = self.unpack(toolchain)
        build = self.make_build_path(toolchain)

        if toolchain.is_arm:
            arch = 'arm'
        elif toolchain.is_aarch64:
            arch = 'aarch64'
        else:
            arch = 'x86'

        if toolchain.is_windows:
            target_os = 'mingw32'
        else:
            target_os = 'linux'

        configure = [
            os.path.join(src, 'configure'),
            '--cc=' + toolchain.cc,
            '--cxx=' + toolchain.cxx,
            '--nm=' + toolchain.nm,
            '--extra-cflags=' + toolchain.cflags + ' ' + toolchain.cppflags + ' ' + self.cppflags,
            '--extra-cxxflags=' + toolchain.cxxflags + ' ' + toolchain.cppflags + ' ' + self.cppflags,
            '--extra-ldflags=' + toolchain.ldflags,
            '--extra-libs=' + toolchain.libs,
            '--ar=' + toolchain.ar,
            '--ranlib=' + toolchain.ranlib,
            '--enable-cross-compile',
            '--arch=' + arch,
            '--target-os=' + target_os,
            '--prefix=' + toolchain.install_prefix,
        ] + self.configure_args

        if toolchain.is_armv7:
            configure.append('--cpu=cortex-a8')

        subprocess.check_call(configure, cwd=build, env=toolchain.env)
        subprocess.check_call(['/usr/bin/make', '--quiet', '-j12'], cwd=build, env=toolchain.env)
        subprocess.check_call(['/usr/bin/make', '--quiet', 'install'], cwd=build, env=toolchain.env)



import subprocess, multiprocessing
from typing import Optional, Sequence, Union

from build.project import Project
from .toolchain import AnyToolchain

class MakeProject(Project):
    def __init__(self, url: Union[str, Sequence[str]], md5: str, installed: str,
                 install_target: str='install',
                 **kwargs):
        Project.__init__(self, url, md5, installed, **kwargs)
        self.install_target = install_target

    def get_simultaneous_jobs(self) -> int:
        try:
            # use twice as many simultaneous jobs as we have CPU cores
            return multiprocessing.cpu_count() * 2
        except NotImplementedError:
            # default to 12, if multiprocessing.cpu_count() is not implemented
            return 12

    def get_make_args(self, toolchain: AnyToolchain) -> list[str]:
        return ['--quiet', '-j' + str(self.get_simultaneous_jobs())]

    def get_make_install_args(self, toolchain: AnyToolchain) -> list[str]:
        return ['--quiet', self.install_target]

    def make(self, toolchain: AnyToolchain, wd: str, args: list[str]) -> None:
        subprocess.check_call(['make'] + args,
                              cwd=wd, env=toolchain.env)

    def build_make(self, toolchain: AnyToolchain, wd: str, install: bool=True) -> None:
        self.make(toolchain, wd, self.get_make_args(toolchain))
        if install:
            self.make(toolchain, wd, self.get_make_install_args(toolchain))



from typing import Sequence, Union
import os
import sys
from tempfile import NamedTemporaryFile
import urllib.request

from .verify import verify_file_digest

def __to_string_sequence(x: Union[str, Sequence[str]]) -> Sequence[str]:
    if isinstance(x, str):
        return (x,)
    else:
        return x

def __get_any(x: Union[str, Sequence[str]]) -> str:
    if isinstance(x, str):
        return x
    else:
        return x[0]

def __download_one(url: str, path: str) -> None:
    print("download", url)
    urllib.request.urlretrieve(url, path)

def __download(urls: Sequence[str], path: str) -> None:
    for url in urls[:-1]:
        try:
            __download_one(url, path)
            return
        except:
            print("download error:", sys.exc_info()[0])
    __download_one(urls[-1], path)

def __download_and_verify_to(urls: Sequence[str], md5: str, path: str) -> None:
    __download(urls, path)
    if not verify_file_digest(path, md5):
        raise RuntimeError("Digest mismatch")

def download_basename(urls: Union[str, Sequence[str]]) -> str:
    return os.path.basename(__get_any(urls))

def download_and_verify(urls: Union[str, Sequence[str]], md5: str, parent_path: str) -> str:
    """Download a file, verify its MD5 checksum and return the local path."""

    base = download_basename(urls)

    os.makedirs(parent_path, exist_ok=True)
    path = os.path.join(parent_path, base)

    try:
        if verify_file_digest(path, md5): return path
        os.unlink(path)
    except FileNotFoundError:
        pass

    with NamedTemporaryFile(dir=parent_path) as tmp:
        __download_and_verify_to(__to_string_sequence(urls), md5, tmp.name)
        os.link(tmp.name, path)

    return path



import os
import shutil
import subprocess
import platform

from .toolchain import AnyToolchain

def format_meson_cross_file_command(command: str) -> str:
    splitted = command.split()
    if len(splitted) == 1:
        return repr(command)

    return repr(splitted)

def make_cross_file(toolchain: AnyToolchain) -> str:
    if toolchain.is_windows:
        system = 'windows'
        windres = "windres = '%s'" % toolchain.windres
    else:
        system = 'linux'
        windres = ''

    if toolchain.is_arm:
        cpu_family = 'arm'
        if toolchain.is_armv7:
            cpu = 'armv7'
        else:
            cpu = 'armv6'
    elif toolchain.is_aarch64:
        cpu_family = 'aarch64'
        cpu = 'arm64-v8a'
    else:
        cpu_family = 'x86'
        if 'x86_64' in toolchain.host_triplet:
            cpu = 'x86_64'
        else:
            cpu = 'i686'

    # TODO: support more CPUs
    endian = 'little'

    # TODO: write pkg-config wrapper

    path = os.path.join(toolchain.build_path, 'meson.cross')
    os.makedirs(toolchain.build_path, exist_ok=True)
    with open(path, 'w') as f:
        f.write(f"""
[binaries]
c = {format_meson_cross_file_command(toolchain.cc)}
cpp = {format_meson_cross_file_command(toolchain.cxx)}
ar = {format_meson_cross_file_command(toolchain.ar)}
strip = {format_meson_cross_file_command(toolchain.strip)}
pkgconfig = {format_meson_cross_file_command(toolchain.pkg_config)}
""")

        if toolchain.is_windows and platform.system() != 'Windows':
            f.write(f"windres = {format_meson_cross_file_command(toolchain.windres)}\n")

            if shutil.which('wine') is not None:
                # Run unit tests with WINE when cross-building for Windows
                print("exe_wrapper = 'wine'", file=f)

        f.write(f"""
[properties]
root = '{toolchain.install_prefix}'
""")

        if toolchain.is_android:
            f.write("""
# Keep Meson from executing Android-x86 test binariees
needs_exe_wrapper = true
""")

        f.write(f"""
[built-in options]
c_args = {repr((toolchain.cppflags + ' ' + toolchain.cflags).split())}
c_link_args = {repr(toolchain.ldflags.split() + toolchain.libs.split())}

cpp_args = {repr((toolchain.cppflags + ' ' + toolchain.cxxflags).split())}
cpp_link_args = {repr(toolchain.ldflags.split() + toolchain.libs.split())}
""")

        f.write(f"""
[host_machine]
system = '{system}'
cpu_family = '{cpu_family}'
cpu = '{cpu}'
endian = '{endian}'
""")
    return path

def configure(toolchain: AnyToolchain, src: str, build: str, args: list[str]=[]) -> None:
    configure = [
        'meson', 'setup',
        build, src,

        '--prefix', toolchain.install_prefix,

        '--buildtype', 'plain',

        '--default-library=static',
    ] + args

    if toolchain.host_triplet is not None:
        # cross-compiling: write a cross-file
        cross_file = make_cross_file(toolchain)
        configure.append(f'--cross-file={cross_file}')

    env = toolchain.env.copy()

    subprocess.check_call(configure, env=env)



#!/usr/bin/env -S python3 -u

import os, os.path
import sys, subprocess
import shutil

configure_args = sys.argv[1:]

x64 = True

while len(configure_args) > 0:
    arg = configure_args[0]
    if arg == '--64':
        x64 = True
    elif arg == '--32':
        x64 = False
    else:
        break
    configure_args.pop(0)

if x64:
    host_arch = 'x86_64-w64-mingw32'
else:
    host_arch = 'i686-w64-mingw32'

# the path to the MPD sources
mpd_path = os.path.abspath(os.path.join(os.path.dirname(sys.argv[0]) or '.', '..'))
sys.path[0] = os.path.join(mpd_path, 'python')

# output directories
from build.dirs import lib_path, tarball_path, src_path
from build.toolchain import MingwToolchain

arch_path = os.path.join(lib_path, host_arch)
build_path = os.path.join(arch_path, 'build')
root_path = os.path.join(arch_path, 'root')

# a list of third-party libraries to be used by MPD on Android
from build.libs import *
thirdparty_libs = [
    zlib,
    libmodplug,
    libopenmpt,
    wildmidi,
    gme,
    ffmpeg,
    libnfs,
    libsamplerate,
]

# build the third-party libraries
toolchain = MingwToolchain(mpd_path,
                           '/usr', host_arch, x64,
                           tarball_path, src_path, build_path, root_path)

for x in thirdparty_libs:
    if not x.is_installed(toolchain):
        x.build(toolchain)

# configure and build MPD

from build.meson import configure as run_meson
run_meson(toolchain, mpd_path, '.', configure_args)
subprocess.check_call(['/usr/bin/ninja'], env=toolchain.env)
